/*
 * Copyright (c) 1997-2006 Erez Zadok,
 * Copyright (c) 2004-2006 Nikolai Joukov,
 * Copyright (c) 2005      Arun M. Krishnakumar,
 * Copyright (c) 2006      Chaitanya Patti,
 * Copyright (c) 2005      Sunil Satnur,
 * Copyright (c) 2006      Avishay Traeger,
 * Copyright (c) 2001-2006 Stony Brook University,
 * Copyright (c) 1997-2000 Columbia University
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: file.c,v 1.89 2006/12/30 22:08:37 crpatti Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "raif.h"
#include <linux/dcache.h>

#if defined(AIO) || defined(AIO_PARITY)
#include <linux/uio.h>
#endif

extern int raif_rule_add(dentry_t *, char *, int);
extern int raif_rule_list(dentry_t *, struct raif_rulebuf *);
extern int raif_rule_del(dentry_t *, int, int);
extern int raif_migrate(file_t *);

/* This is not meant to be a generic repositioning function.  If you do
 * things that aren't supported, then we return EINVAL.
 *
 * What is allowed:
 *  (1) seeking to the same position that you are currently at
 *	This really has no effect, but returns where you are.
 *  (2) seeking to the end of the file, if you've read everything
 *	This really has no effect, but returns where you are.
 *  (3) seeking to the beginning of the file
 *	This throws out all state, and lets you begin again.
 */
STATIC loff_t raif_dir_llseek(struct file *file, loff_t offset, int origin)
{
	struct raif_dir_state *rdstate;
	loff_t err = 0;

	print_entry(" file=%p, offset=0x%llx, origin = %d", file, offset,
		    origin);

	/* ArunMK
	if ((err = raif_file_revalidate(file, 0)))
		goto out;
	*/
	rdstate = ftopd(file)->rdstate;

	/* We let users seek to their current position, but not anywhere else. */
	if (!offset) {
		switch (origin) {
		case SEEK_SET:
			if (rdstate) {
				free_rdstate(rdstate);
				ftopd(file)->rdstate = NULL;
			}
			init_rdstate(file);
			err = 0;
			break;
		case SEEK_CUR:
			if (file->f_pos) {
				if (file->f_pos == DIREOF)
					err = DIREOF;
				else
					ASSERT(file->f_pos ==
					       rdstate2offset(rdstate));
				err = file->f_pos;
			} else {
				err = 0;
			}
			break;
		case SEEK_END:
			/* Unsupported, because we would break everything.  */
			err = -EINVAL;
			break;
		}
	} else {
		switch (origin) {
		case SEEK_SET:
			if (rdstate) {
				if (offset == rdstate2offset(rdstate)) {
					err = offset;
				} else if (file->f_pos == DIREOF) {
					err = DIREOF;
				} else {
					err = -EINVAL;
				}
			} else {
				if ((rdstate =
				     find_rdstate(file->f_dentry->d_inode,
						  offset))) {
					ftopd(file)->rdstate = rdstate;
					err = rdstate->uds_offset;
				} else {
					err = -EINVAL;
				}
			}
			break;
		case SEEK_CUR:
		case SEEK_END:
			/* Unsupported, because we would break everything.  */
			err = -EINVAL;
			break;
		}
	}

	print_exit_status((int)err);
	return err;
}




/* Make sure our rdstate is playing by the rules. */
static void verify_rdstate_offset(struct raif_dir_state *rdstate)
{
	PASSERT(rdstate);
	ASSERT(rdstate->uds_offset < DIREOF);
	ASSERT(rdstate->uds_cookie < MAXRDCOOKIE);
}

struct raif_getdents_callback {
	struct raif_dir_state *rdstate;
	void *dirent;
	int entries_written;
	int filldir_called;
	int filldir_error;
	filldir_t filldir;
	struct super_block *sb;
};

/* copied from generic filldir in fs/readir.c */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
STATIC int raif_filldir(void *dirent, const char *name, int namelen,
			   loff_t offset, ino_t ino, unsigned int d_type)
#else
STATIC int raif_filldir(void *dirent, const char *name, int namelen,
			   loff_t offset, u64 ino, unsigned int d_type)
#endif
{
	struct raif_getdents_callback *buf =
	    (struct raif_getdents_callback *)dirent;
	struct filldir_node *found = NULL;
	int err = 0;
	off_t pos = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
	ino_t raif_ino;
#else
	u64 raif_ino;
#endif

	fist_dprint(9, "raif_filldir name=%*s\n", namelen, name);

	PASSERT(buf);
	buf->filldir_called++;

	//do not show the rules files to the user in the "ls" command
	if (!strncmp(name, RAIF_RULES_FILE_NAME, namelen) && (namelen == RAIF_RULES_FILE_NAME_LEN)) {
		goto out;
	}
	found = find_filldir_node(buf->rdstate, name, namelen);

	if (found)
		goto out;

	pos = rdstate2offset(buf->rdstate);
	raif_ino = ino;

	err = buf->filldir(buf->dirent, name, namelen, pos,
			   raif_ino, d_type);
	buf->rdstate->uds_offset++;
	verify_rdstate_offset(buf->rdstate);

	/* If we did fill it, stuff it in our hash, otherwise return an error */
	if (err) {
		buf->filldir_error = err;
		goto out;
	}
	buf->entries_written++;
	if ((err = add_filldir_node(buf->rdstate, name, namelen, buf->rdstate->uds_bindex)))
		buf->filldir_error = err;

out:
	return err;
}

STATIC int raif_readdir_fast(file_t *file, void *dirent, filldir_t filldir)
{
	int err = -ENOTDIR;
	file_t *hidden_file;

        print_entry_location();

	hidden_file = ftohf_index(file, 0);
	hidden_file->f_pos = file->f_pos;
	err = vfs_readdir(hidden_file, filldir, dirent);
	file->f_pos = hidden_file->f_pos;
	fist_copy_attr_atime(file->f_dentry->d_inode, hidden_file->f_dentry->d_inode);
	
	print_exit_status(err);
	
	return err;
}											

STATIC int raif_readdir(file_t *file, void *dirent, filldir_t filldir)
{
	int err = -ENOTDIR;
	file_t *hidden_file;
	int bnum = 0;
	struct raif_getdents_callback buf;
	struct raif_dir_state *uds;
	struct inode *inode = NULL;
	loff_t offset;

	print_entry_location();

	PASSERT(file);
	PASSERT(file->f_dentry);
	PASSERT(file->f_dentry->d_inode);
	inode = file->f_dentry->d_inode;

	uds = ftopd(file)->rdstate;
	if (!uds) {
		if (file->f_pos == DIREOF) {
			err = 0;
			goto out;
		} else if (file->f_pos > 0) {
			uds = find_rdstate(inode, file->f_pos);
			if (!uds) {
				err = -ESTALE;
				goto out;
			}
			ftopd(file)->rdstate = uds;
		} else {
			init_rdstate(file);
			uds = ftopd(file)->rdstate;
		}
	}
	bnum = fbnum(file);

	/* have to continue from same branch as this branch may not
	   have been completely read in the previous run */
	uds->uds_bindex = GetNextBranch(uds->minSet, uds->uds_bindex);

	while (uds->uds_bindex < bnum) {
		hidden_file = ftohf_index(file, uds->uds_bindex);
		if (!hidden_file) {
			fist_dprint(7,
				    "Incremented bindex to %d of %d,"
				    " because hidden file is NULL.\n",
				    uds->uds_bindex, bnum);
			uds->uds_bindex++;
			uds->uds_bindex = GetNextBranch(uds->minSet, uds->uds_bindex);
			uds->uds_dirpos = 0;
			continue;
		}

		/* prepare callback buffer */
		buf.filldir_called = 0;
		buf.filldir_error = 0;
		buf.entries_written = 0;
		buf.dirent = dirent;
		buf.filldir = filldir;
		buf.rdstate = uds;
		buf.sb = inode->i_sb;

		/* Read starting from where we last left off. */
		offset = vfs_llseek(hidden_file, uds->uds_dirpos, 0);
		if (offset < 0) {
			err = offset;
			goto out;
		}
		fist_dprint(7, "calling readdir for %d.%lld (offset = %lld)\n",
			    uds->uds_bindex, uds->uds_dirpos, offset);
		err = vfs_readdir(hidden_file, raif_filldir, (void *)&buf);
		fist_dprint(7,
			    "readdir on %d.%lld = %d (entries written %d, filldir called %d)\n",
			    uds->uds_bindex, (long long)uds->uds_dirpos, err,
			    buf.entries_written, buf.filldir_called);

		/* Save the position for when we continue. */
		offset = vfs_llseek(hidden_file, 0, 1);
		if (offset < 0) {
			err = offset;
			goto out;
		}
		uds->uds_dirpos = offset;

		/* Copy the atime. */
		fist_copy_attr_atime(inode, hidden_file->f_dentry->d_inode);

		if (err < 0) {
			goto out;
		}

		if (buf.filldir_error) {
			break;
		}

		if (!buf.entries_written) {
			uds->uds_bindex++;
			uds->uds_bindex = GetNextBranch(uds->minSet, uds->uds_bindex);
			uds->uds_dirpos = 0;
		}
	}

	if (!buf.filldir_error && uds->uds_bindex >= bnum) {
		fist_dprint(3, "Discarding rdstate because readdir is over (hashsize = %d)\n",
			    uds->uds_hashentries);
		/* Save the number of hash entries for next time. */
		itopd(inode)->uii_hashsize = uds->uds_hashentries;
		free_rdstate(uds);
		ftopd(file)->rdstate = NULL;
		file->f_pos = DIREOF;
	} else {
		PASSERT(ftopd(file)->rdstate);
		file->f_pos = rdstate2offset(uds);
		fist_dprint(3, "rdstate now has a cookie of %u (err = %d)\n",
			    uds->uds_cookie, err);
	}

      out:
	print_exit_status(err);
	return err;
}

STATIC unsigned int raif_poll(file_t *file, poll_table *wait)
{
/*	unsigned int mask = DEFAULT_POLLMASK;
	file_t *hidden_file = NULL;

	print_entry_location();
	if (ftopd(file) != NULL)
		hidden_file = ftohf_index(file, 0);

	if (!hidden_file->f_op || !hidden_file->f_op->poll)
		goto out;

	mask = hidden_file->f_op->poll(hidden_file, wait);

out:
	print_exit_status(mask);
		return mask;*/
	return -ENODEV;
}

/* minimal rule for addition is "G:0:*:1:0:0" */
#define MINIMAL_RULE_LENGTH 12

STATIC inline char* raif_rule_copy(char __user *userRule)
{
	int err = 0, len;
	char* kernRule;

	len = strlen_user((char *)userRule);
	/* catch errors as soon as possible */
	if (len < MINIMAL_RULE_LENGTH) {
		err = -EINVAL;
		printk(KERN_ERR "RAIF_IOCTL: Invalid rule string length\n");
		goto out_error;
	}
	
	kernRule = KMALLOC(len, GFP_KERNEL);
	if (!kernRule) {
		err = -ENOMEM;
		printk(KERN_ERR "RAIF_IOCTL: No memory to allocate rule string\n");
		goto out_error;
	}
	
	if (copy_from_user(kernRule, userRule, len)) {
		err = -EINVAL;
		printk(KERN_ERR "RAIF_IOCTL: Unable to copy rule string from user-space\n");
		goto out_free;
	}
	return kernRule;
out_free:
	KFREE(kernRule);
out_error:
	return ERR_PTR(err);
}


STATIC int raif_ioctl(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg)
{
	int err = 0;
	file_t *hidden_file = NULL;
	vfs_t *this_vfs;
	vnode_t *this_vnode;
	int val;
	char* rule;
	struct raif_rulebuf *rulebuf;

	print_entry_location();

	this_vfs = inode->i_sb;
	this_vnode = inode;

	switch (cmd) {
		case FIST_IOCTL_GET_DEBUG_VALUE:
			val = fist_get_debug_value();
			printk("IOCTL GET: send arg %d\n", val);
			err = put_user(val, (int *)arg);
			break;

		case FIST_IOCTL_SET_DEBUG_VALUE:
			err = get_user(val, (int *)arg);
			if (err)
				break;
			fist_dprint(6, "IOCTL SET: got arg %d\n", val);
			if (val < 0 || val > 20) {
				err = -EINVAL;
				break;
			}
			fist_set_debug_value(val);
			break;

		case RAIF_IOCTL_RULE_ADD:
			rule = raif_rule_copy((char *) arg);
			if (IS_ERR(rule)) {
				err = PTR_ERR(rule);
				break;
			}
			err = raif_rule_add(file->f_dentry, rule, 1);
			KFREE(rule);
			break;
		case RAIF_IOCTL_RULE_DEL:
			rulebuf = (struct raif_rulebuf *)arg;
                        if (!access_ok(VERIFY_READ, rulebuf, sizeof(struct raif_rulebuf))) {
                                printk(KERN_ERR "RAIF IOCTL: Invalid user-space data block\n");
                                err = -EFAULT;
                                break;
                        }
			err = raif_rule_del(file->f_dentry, rulebuf->ruletype, rulebuf->rulenum);
			break;
		case RAIF_IOCTL_RULE_LIST:
                        rulebuf = (struct raif_rulebuf *)arg;
                        if (!access_ok(VERIFY_WRITE, rulebuf, sizeof(struct raif_rulebuf))) {
                                printk(KERN_ERR "RAIF IOCTL: Invalid user-space data block\n");
                                err = -EFAULT;
                                break;
                        }
                        if (!access_ok(VERIFY_WRITE, rulebuf->buf, RAIF_RULEBUF_SIZE)) {
                                printk(KERN_ERR "RAIF IOCTL: Invalid user-space rule buffer\n");
                                err = -EFAULT;
                                break;
                        }
                        err = raif_rule_list(file->f_dentry, rulebuf);
                        break;

                case RAIF_IOCTL_MIGRATE:
                        err = raif_migrate(file);
                        break;

			/* add non-debugging fist ioctl's here */
		default:
			if (ftopd(file) != NULL)
				hidden_file = ftohf_index(file, 0);
			/* pass operation to hidden filesystem, and return status */
			if (hidden_file && hidden_file->f_op && hidden_file->f_op->ioctl)
				err = hidden_file->f_op->ioctl(itohi_index(inode, 0), hidden_file, cmd, arg);
			else
				err = -ENOTTY;	/* this is an unknown ioctl */
	} /* end of outer switch statement */
	print_exit_status(err);
	return err;
}

STATIC int raif_open(inode_t *inode, file_t *file)
{
	int err = -ENOMEM;
	int hidden_flags;
	file_t *hidden_file = NULL;
	dentry_t *hidden_dentry = NULL;
	dentry_t *dentry = NULL;
	super_block_t *sb;
	int bindex = 0, bnum = 0, global_bindex;

	print_entry_location();
	//fist_dprint(8, "    open: %s\n", file->f_dentry->d_name.name);
	/* ftopd(file) */ 
	file->private_data = KMALLOC(sizeof(struct raif_file_info), GFP_KERNEL);
	if (!ftopd(file))
		goto out;

	dentry = file->f_dentry;
	init_ftopd(file);
	bnum = fbnum(file) = dbnum(dentry);

	ftohf_ptr(file) = KMALLOC((sizeof(file_t *) * bnum), GFP_KERNEL);
	if (!ftohf_ptr(file))
		goto out;
	init_ftohf_ptr(file, bnum);
	hidden_flags = file->f_flags;
	/* have to handle files opened for append specially because 
	 * position passed to raif_file_write or raif_file_read will not be
	 * set properly
	 */
	hidden_flags &= ~(O_APPEND);
	if (hidden_flags & O_WRONLY) {
		hidden_flags &= ~(O_WRONLY);
		hidden_flags |= O_RDWR;
	}
	PASSERT(dentry);

	sb = inode->i_sb;
	err = 0;
	for (bindex = 0; bindex < bnum; bindex++) {

		global_bindex = GLOBAL_BINDEX(dentry, bindex);

		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry)
			continue;

		dget(hidden_dentry);
		/*
		 * dentry_open will decrement mnt refcnt if err.
		 * otherwise fput() will do an mntput() for us upon file close.
		 */
		mntget(stohm_index(sb, global_bindex));
		PASSERT(hidden_dentry);
		if (!hidden_dentry->d_inode) {
			printk("<0> negative dentry is [%s] at [%d]\n", dentry->d_name.name, bindex);
		}
		PASSERT(hidden_dentry->d_inode);
		hidden_file = dentry_open(hidden_dentry, stohm_index(sb, global_bindex), hidden_flags);
		if (IS_ERR(hidden_file)) {
			err =  PTR_ERR(hidden_file);
			goto out;
		}
		ftohf_index(file, bindex) = hidden_file;	/* link two files */
	}

out:
	/* freeing the allocated resources, and fput the opened files */
	if (err < 0 && ftopd(file)) {
		for (; bindex >= 0; bindex--) {
			hidden_file  = ftohf_index(file, bindex);
			if (hidden_file) {
				/* fput calls dput for hidden_dentry */
				fput(hidden_file);  
			}
		}
		if (ftohf_ptr(file)) {
			KFREE(ftohf_ptr(file));
		}
		KFREE(ftopd(file));
	}

	print_exit_status(err);
	return err;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
STATIC int raif_flush(file_t *file)
#else
STATIC int raif_flush(file_t *file, fl_owner_t id)
#endif
{
	int err = 0;
	file_t *hidden_file = NULL;
	int bindex;

	print_entry_location();
	if (ftopd(file) != NULL) { 
		for (bindex = 0; bindex < fbnum(file); bindex++) {
			hidden_file = ftohf_index(file, bindex);
			if (!hidden_file || !hidden_file->f_op || !hidden_file->f_op->flush)
				continue;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
			err = hidden_file->f_op->flush(hidden_file);
#else
			err = hidden_file->f_op->flush(hidden_file, id);
#endif
		}
	}
	print_exit_status(err);
	return err;
}


STATIC int raif_release(inode_t *inode, file_t *file)
{
	int err = 0;
	file_t *hidden_file = NULL;
	int bindex, bnum;

	print_entry_location();

	/*finish pending writes*/
	if (stopd(file->f_dentry->d_sb)->is_lower_nfs) {
		if (file->f_mode & FMODE_WRITE)
			filemap_fdatawrite(file->f_mapping);
	}

	/*fist_checkinode(inode, "raif_release");*/
	/* fput all the hidden files */
	bnum = fbnum(file);
	for (bindex = 0; bindex < bnum; bindex++) {
		hidden_file = ftohf_index(file, bindex);
		if (hidden_file) {
			/* will decrement file refcount, and if 0, destroy the file,
			 * which will call the lower file system's file release function. */
			fput(hidden_file);
		}
	}
	KFREE(ftohf_ptr(file));
	KFREE(ftopd(file));

	/*fist_checkinode(inode, "post raif_release");*/
	print_exit_status(err);
	return err;
}

STATIC int raif_fsync(file_t *file, dentry_t *dentry, int datasync)
{
	int err = -EINVAL;
	file_t *hidden_file = NULL;
	dentry_t *hidden_dentry;
	int bindex;
	print_entry_location();

	if (ftopd(file) != NULL) { 
		for (bindex = 0; bindex < fbnum(file); bindex++) {
			hidden_file = ftohf_index(file, bindex);
			hidden_dentry = dtohd_index(dentry, bindex);
			if (hidden_file->f_op && hidden_file->f_op->fsync) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
				down(&hidden_dentry->d_inode->i_sem);
#else
				mutex_lock(&hidden_dentry->d_inode->i_mutex);
#endif
				err = hidden_file->f_op->fsync(hidden_file, hidden_dentry, datasync);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
				up(&hidden_dentry->d_inode->i_sem);
#else
				mutex_unlock(&hidden_dentry->d_inode->i_mutex);
#endif
			}
		}
	}

	print_exit_status(err);
	return err;
}


STATIC int raif_fasync(int fd, file_t *file, int flag)
{
	int err = 0;
	file_t *hidden_file = NULL;
	int bindex;

	print_entry_location();

	if (ftopd(file) != NULL) { 
		for (bindex = 0; bindex < fbnum(file); bindex++) {
			hidden_file = ftohf_index(file, bindex);
			err = hidden_file->f_op->fasync(fd, hidden_file, flag);
		}
	}
	print_exit_status(err);
	return err;
}


STATIC int raif_lock(file_t *file, int cmd, struct file_lock *fl)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
	struct file_lock conflock;
#endif
	int err = 0;
	file_t *hidden_file = NULL;
	int bindex;

	print_entry_location();
	if (ftopd(file) != NULL) { 
		for (bindex = 0; bindex < fbnum(file); bindex++) {
			hidden_file = ftohf_index(file, bindex);
			if (hidden_file->f_op && hidden_file->f_op->lock)
				err = hidden_file->f_op->lock(hidden_file, F_GETLK, fl);
			else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
				posix_test_lock(hidden_file, fl);
#else
				posix_test_lock(hidden_file, fl, &conflock);
#endif
		}
	}

	print_exit_status(err);
	return err;
}

/* Trimmed directory options, we shouldn't pass everything down since
 * we don't want to operate on partial directories.
 */
struct file_operations raif_dir_fops =
{
	llseek:		raif_dir_llseek,
	read:		generic_read_dir, /* always returns error */
#if defined(RAIF_TAG) && !defined(RAIF_TAG_PAGE)
	readdir:	raif_readdir,
#else
	readdir:	raif_readdir_fast,
#endif
	ioctl:		raif_ioctl,
	open:		raif_open,
	release:	raif_release,
	flush:		raif_flush,
};

struct file_operations raif_main_fops =
{
	llseek:		generic_file_llseek,

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
	read:		generic_file_read,
	write:		generic_file_write,
#else
	read:		do_sync_read,
	aio_read:	generic_file_aio_read,
	write:		do_sync_write,
	aio_write:	generic_file_aio_write,
#endif

#if defined(RAIF_TAG) && !defined(RAIF_TAG_PAGE)
	readdir:	raif_readdir,
#else
	readdir:	raif_readdir_fast,
#endif
	poll:		raif_poll,
	ioctl:		raif_ioctl,
	mmap:		generic_file_mmap,
	open:		raif_open,
	flush:		raif_flush,
	release:	raif_release,
	fsync:		raif_fsync,
	fasync:		raif_fasync,
	lock:		raif_lock,
	/* not needed: readv */
	/* not needed: writev */
	/* not implemented: sendpage */
	/* not implemented: get_unmapped_area */
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
