/*
 * Copyright (c) 1997-2006 Erez Zadok,
 * Copyright (c) 2004-2006 Nikolai Joukov,
 * Copyright (c) 2005      Arun M. Krishnakumar,
 * Copyright (c) 2006      Chaitanya Patti,
 * Copyright (c) 2006      Avishay Traeger,
 * Copyright (c) 2001-2006 Stony Brook University,
 * Copyright (c) 1997-2000 Columbia University
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: inode.c,v 1.98 2007/01/03 02:19:25 crpatti Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "raif.h"
#include "ECC/levels.h"


// put the first dentry into the right place
#define DENTRY0_TO_GLOBAL \
		if (bindex == 0) {\
			global_bindex = GLOBAL_BINDEX(dentry, 0); \
			if (0 != global_bindex && !dtohd_index(dentry, global_bindex)) { \
				dtohd_index(dentry, global_bindex) = hidden_dentry; \
				hidden_dentry = 0; \
			} \
		}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)

STATIC dentry_t * raif_cached_lookup(dentry_t * parent, struct qstr * name, struct nameidata *nd)
{
	struct dentry * dentry /*= __d_lookup(parent, name)*/;

	/* lockess __d_lookup may fail due to concurrent d_move() 
	 * in some unrelated directory, so try with d_lookup
	 */
	/*if (!dentry)*/
		dentry = d_lookup(parent, name);

	if (dentry && dentry->d_op && dentry->d_op->d_revalidate) {
		int status = dentry->d_op->d_revalidate(dentry, nd);
		if (status <= 0) {
			if (!status) {
				if (!d_invalidate(dentry)) {
					dput(dentry);
					dentry = NULL;
				} 
			} else {
				dput(dentry);
				dentry = ERR_PTR(status);
			}
		}
	}

	return dentry;
}

STATIC dentry_t * raif_lookup_hash(struct qstr *name, dentry_t * base, struct nameidata *nd)
{
	struct dentry * dentry;
	struct inode *inode;
	int err;

	inode = base->d_inode;
	err = permission(inode, MAY_EXEC, nd);
	dentry = ERR_PTR(err);
	if (err)
		goto out;

	/*
	 * See if the low-level filesystem might want
	 * to use its own hash..
	 */
	if (base->d_op && base->d_op->d_hash) {
		err = base->d_op->d_hash(base, name);
		dentry = ERR_PTR(err);
		if (err < 0)
			goto out;
	}

	dentry = raif_cached_lookup(base, name, nd);
	if (!dentry) {
		struct dentry *new = d_alloc(base, name);
		dentry = ERR_PTR(-ENOMEM);
		if (!new)
			goto out;
		dentry = inode->i_op->lookup(inode, new, nd);
		if (!dentry)
			dentry = new;
		else
			dput(new);
	}
out:
	return dentry;
}

STATIC inline dentry_t *raif_lookup_one_len_nd(const char * name, dentry_t * base, int len, struct nameidata * nd)
{
	unsigned long hash;
	struct qstr this;
	unsigned int c;

	this.name = name;
	this.len = len;
	if (!len)
		goto access;

	hash = init_name_hash();
	while(len--) {
		c = *(const unsigned char *)name++;
		if (c == '/' || c == '\0')
			goto access;
		hash = partial_name_hash(c, hash);
	}
	this.hash = end_name_hash(hash);
	
	return raif_lookup_hash(&this, base, nd);

access:
	return ERR_PTR(-EACCES);
}

#endif 

STATIC inline dentry_t* raif_lookup_branch(dentry_t* dentry, int bindex, struct nameidata *nd)
{
/*#ifdef RAIF_LB
	struct timeval b4, af;
#endif*/
	struct nameidata newnd;
	dentry_t* lower_dentry, *hidden_parent;

	print_entry_location();

	hidden_parent = dtohd_index(dentry->d_parent, bindex);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
	if (nd) {
		newnd = *nd;
		newnd.dentry = hidden_parent;
		newnd.mnt = stohm_index(dentry->d_sb, GLOBAL_BINDEX(dentry, bindex)); 
	}

 	lower_dentry = raif_lookup_one_len_nd(dentry->d_name.name, hidden_parent, dentry->d_name.len, (nd?(&newnd):NULL));
#else
	/*LB_UPDATE_PREPARE;*/
	lower_dentry = lookup_one_len(dentry->d_name.name, hidden_parent, dentry->d_name.len);
	/*LB_UPDATE_COMMIT_I(dir, global_bindex, lookup);*/ /*FIXME: change to global_bindex below*/
	/*LB_UPDATE_COMMIT_I(dentry->d_parent->d_inode, bindex, lookup);*/
#endif

	fist_dprint(8, "    IN %s: branch=%d\n", __FUNCTION__, bindex);

	if (IS_ERR(lower_dentry))
		lower_dentry = NULL;
		
	print_exit_status(~(int)lower_dentry);

	return lower_dentry;
}

dentry_t *raif_lookup_backend(inode_t *dir, dentry_t *dentry, int interpose_flag, struct nameidata *nd)
{ 
	int err = 0;
	dentry_t *hidden_dentry = NULL;
	dentry_t *tmp_dentry = NULL;
	int bindex, global_bindex, bnum, seed_bindex = -1;

	print_entry_location();

	fist_dprint(8, "    IN %s: name=\"%s\" flag=%d\n", __FUNCTION__, dentry->d_name.name, interpose_flag);

	dentry->d_op = &raif_dops;

	dentry->d_fsdata = (struct raif_dentry_info *) KMALLOC(sizeof(struct raif_dentry_info), GFP_KERNEL);
	if (!dtopd(dentry)) {
		err = -ENOMEM;
		goto out;
	}
	init_dtopd(dentry, dir->i_sb);
	dtohd_ptr(dentry) = KMALLOC(sizeof(dentry_t *) * ibnum(dir), GFP_KERNEL);
	if (!dtohd_ptr(dentry)) {
		err = -ENOMEM;
		goto out_free_dtopd;
	}
	init_dtohd_ptr(dentry, ibnum(dir));

#if defined(RAIF_TAG) && !defined(RAIF_TAG_PAGE)
	err = raif_rule_lookup(dentry);
	bnum = dbnum(dentry);
#else
	bnum = dbnum(dentry) = ibnum(dir);
#endif


	for (bindex = 0; bindex < bnum; bindex++) {
		int target_bindex = bindex;
#if defined(RAIF_TAG) && !defined(RAIF_TAG_PAGE)
		// bindex > 0 -> d_inode != NULL
		if ((bindex > 0) && tmp_dentry) {
			if (S_ISREG(tmp_dentry->d_inode->i_mode)) {
				global_bindex = GLOBAL_BINDEX(dentry, bindex);
			} else {
				global_bindex = bindex;
				if (seed_bindex == global_bindex) {
					global_bindex = 0;
					target_bindex = 0;
				}
			}
		} else
#endif
		global_bindex = GLOBAL_BINDEX(dentry, bindex);

		hidden_dentry = raif_lookup_branch(dentry, global_bindex, nd);
		if (!hidden_dentry) {
			err = -ENOMEM;
			goto out_free;
		}

		if (hidden_dentry->d_inode) {
			tmp_dentry = hidden_dentry;
		} else {
			tmp_dentry = NULL;
		}


		dtohd_index(dentry, target_bindex) = hidden_dentry;

		if (bindex == 0) {
			if (hidden_dentry->d_inode) {
				seed_bindex = global_bindex;
#if defined(RAIF_TAG) && !defined(RAIF_TAG_PAGE)
				if (!S_ISREG(hidden_dentry->d_inode->i_mode)) {
					bnum = dbnum(dentry) = ibnum(dir);
					dtohd_index(dentry, bindex) = NULL;
					dtohd_index(dentry, global_bindex) = hidden_dentry;

					if (dentry != dentry->d_parent) {
						//this will anyway be interposed below
						dtopd(dentry)->d.rules.rule = dtopd(dentry)->d.rules.local_rule = NULL;
					}
				}
#endif
			} else {
				// do not lookup more than one branch if the file doesn't exist
				bnum = 1;
			}
		}
	}

	dir->i_atime = CURRENT_TIME;

	/* If we have found a positive dentry. */ 
	if (seed_bindex >= 0) {
		err = raif_interpose(dentry, dir->i_sb, interpose_flag);
		if (err)
			goto out_free;
	} else {
		if (interpose_flag == INTERPOSE_REVAL) {
		} else {
			d_add(dentry, NULL);
		}
	}
	goto out;

out_free:    
	/* should dput all the underlying dentries on error condition */
	for (bindex = 0; bindex < bnum; bindex++) {
		hidden_dentry = dtohd_index(dentry, bindex);
		if (hidden_dentry)
			dput(hidden_dentry);
	}
	KFREE(dtohd_ptr(dentry));
	dtohd_ptr(dentry) = NULL;
	d_drop(dentry);

out_free_dtopd:
	KFREE(dtopd(dentry));
	/* dtopd(dentry) */
	dentry->d_fsdata = NULL;
out:
	fist_dprint(8, "    OUT d_entry: %p, d_inode: %p\n", dentry, dentry->d_inode);
	print_exit_status(err);
	return ERR_PTR(err);
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
dentry_t * raif_lookup(inode_t *dir, dentry_t *dentry) {
#else
dentry_t * raif_lookup(inode_t *dir, dentry_t *dentry, struct nameidata* nd) {
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	return raif_lookup_backend(dir, dentry, INTERPOSE_LOOKUP, NULL);
#else
	return raif_lookup_backend(dir, dentry, INTERPOSE_LOOKUP, nd);
#endif
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
STATIC int raif_create(inode_t *dir, dentry_t *dentry, int mode)
#else
STATIC int raif_create(inode_t *dir, dentry_t *dentry, int mode, struct nameidata *nd)
#endif
{
	int err = 0, orig_err = 0;
	dentry_t *hidden_dentry = NULL;
	dentry_t *hidden_dir_dentry = NULL;
	int bnum, bindex, ubound;

	print_entry_location();

	// Should get a negative dentry. Create on all branches.
	ASSERT(dentry->d_inode == NULL);

	fist_dprint(8, "name: %s\n", dentry->d_name.name);

	bnum = dbnum(dentry);

	ubound = bnum;
	

	for (bindex = 0; bindex < ubound; bindex++) {	    
		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry) {
			dtohd_index(dentry, bindex) = hidden_dentry = raif_lookup_branch(dentry, GLOBAL_BINDEX(dentry, bindex), nd);
		}
		PASSERT(hidden_dentry);

		hidden_dir_dentry = lock_parent(hidden_dentry);
		if (IS_ERR(hidden_dir_dentry)) {
			err = PTR_ERR(hidden_dir_dentry);
			goto dec_bindex;
		}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		err = vfs_create(hidden_dir_dentry->d_inode, hidden_dentry, mode);
#else
		err = vfs_create(hidden_dir_dentry->d_inode, hidden_dentry, mode, nd);
#endif
		/* XXX this could potentially return a negative hidden_dentry! */
		if (err || !hidden_dentry->d_inode) {
			unlock_dir(hidden_dir_dentry);
			goto error;
		}

		unlock_dir(hidden_dir_dentry);
	} // end for loop

	/* this will create the tag info */
	err = raif_interpose(dentry, dir->i_sb, INTERPOSE_DEFAULT);
	if (err)
		goto error;
		
	/* update parent dir */
	fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
	/* update number of links on parent directory */
	dir->i_nlink = get_nlinks(dir);

	goto end;

dec_bindex:
	bindex--;

error:
	orig_err = err;
	// UNDO code
	// start unlinking files from bindex to 0
	for (; bindex >= 0; bindex--) {

		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry || !hidden_dentry->d_inode)
			continue;

		hidden_dir_dentry = lock_parent(hidden_dentry);

		/* avoid destroying the hidden inode if the file is in use */
		dget(hidden_dentry);
		err = vfs_unlink(hidden_dir_dentry->d_inode, hidden_dentry);
		dput(hidden_dentry);
		fist_copy_attr_times(dir, hidden_dir_dentry->d_inode);
		if (err) {
			unlock_dir(hidden_dir_dentry);
			orig_err = -EIO;
			goto end;
		}

		if (!err) {		/* since it was OK, we kill the hidden dentry. */
			d_delete(hidden_dentry);
		}
		unlock_dir(hidden_dir_dentry);
	}

	err = orig_err;

end:
	print_exit_status(err);
	return err;
}


STATIC int raif_link(dentry_t *old_dentry, inode_t *dir, dentry_t *new_dentry)
{
	int err = 0;
	dentry_t *hidden_old_dentry = NULL;
	dentry_t *hidden_new_dentry = NULL;
	dentry_t *hidden_dir_dentry = NULL;
	dentry_t *hidden_dentry = NULL;
	int bnum, bindex = 0;

	print_entry_location();

	/* not supported yet */
	ASSERT(0);

	bnum = dbnum(new_dentry) = ibnum(dir);
	for (bindex = 0; bindex < bnum; bindex++) {
		hidden_new_dentry = dtohd_index(new_dentry, bindex);
		if (!hidden_new_dentry) {
			dtohd_index(new_dentry, bindex) = hidden_new_dentry = raif_lookup_branch(new_dentry, bindex, NULL);
		}
		PASSERT(hidden_new_dentry);

		hidden_old_dentry = dtohd_index(old_dentry, bindex);

		if (hidden_old_dentry && hidden_new_dentry) {

			dget(hidden_old_dentry);
			hidden_dir_dentry = lock_parent(hidden_new_dentry);

			err = vfs_link(hidden_old_dentry, hidden_dir_dentry->d_inode, hidden_new_dentry);

			if (err || !hidden_new_dentry->d_inode) {
				unlock_dir(hidden_dir_dentry);
				dput(hidden_old_dentry);
				goto error;
			}
			unlock_dir(hidden_dir_dentry);
			dput(hidden_old_dentry);
		}
	} // end for loop	    

	err = raif_interpose(new_dentry, dir->i_sb, INTERPOSE_DEFAULT);
	if (err) {
		goto error;
	}

	fist_copy_attr_timesizes(dir, hidden_new_dentry->d_inode);
	/* propagate number of hard-links */
	old_dentry->d_inode->i_nlink = get_nlinks(old_dentry->d_inode);

	goto end;

error:
	// UNDO code
	// start unlinking files from bindex -> 0
	for (; bindex >= 0; bindex--) {

		hidden_dentry = dtohd_index(new_dentry, bindex);
		if (!hidden_dentry || !hidden_dentry->d_inode)  
			continue;

		hidden_dir_dentry = lock_parent(hidden_dentry);

		/* avoid destroying the hidden inode if the file is in use */
		dget(hidden_dentry);
		err = vfs_unlink(hidden_dir_dentry->d_inode, hidden_dentry);
		dput(hidden_dentry);
		fist_copy_attr_times(dir, hidden_dir_dentry->d_inode);
		if (err) {
			unlock_dir(hidden_dir_dentry);
			goto end;
		}

		if (!err) {		/* since it was OK, we kill the hidden dentry. */
			d_delete(hidden_dentry);
		}
		unlock_dir(hidden_dir_dentry);
	}
end:
	if (!new_dentry->d_inode)
		d_drop(new_dentry);

	print_exit_status(err);
	return err;
}

STATIC int
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
raif_unlink_smallfiles(inode_t *dir, dentry_t *dentry)
#else
raif_unlink_smallfiles(inode_t *dir, dentry_t *dentry, struct nameidata *nd)
#endif
{
	int err = 0, orig_err = 0;
	dentry_t *hidden_dentry;
	dentry_t *hidden_dir_dentry;
	int bindex = 0;
	int mode = 0x666;

	print_entry_location();

	for (bindex = 0; bindex < dbnum(dentry); bindex++) {

		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry)  
			continue;

		hidden_dir_dentry = lock_parent(hidden_dentry);

		/* avoid destroying the hidden inode if the file is in use */
		dget(hidden_dentry);
		err = vfs_unlink(hidden_dir_dentry->d_inode, hidden_dentry);
		if (err) {
			unlock_dir(hidden_dir_dentry);
			goto error;
		} else {
			dput(hidden_dentry);
			d_delete(hidden_dentry);
			fist_copy_attr_times(dir, hidden_dir_dentry->d_inode);
		}

		unlock_dir(hidden_dir_dentry);
	}

	/* propagate number of hard-links */
	dentry->d_inode->i_nlink = get_nlinks(dentry->d_inode);

	goto end;

error:
	orig_err = err;
	for (bindex--; bindex >= 0; bindex--) {
		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry) {
			continue;
		}
		PASSERT(hidden_dentry);
		hidden_dir_dentry = lock_parent(hidden_dentry);
		if (IS_ERR(hidden_dir_dentry)) {
			err = PTR_ERR(hidden_dir_dentry);
			err = -EIO;
			goto end;
		}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		err = vfs_create(hidden_dir_dentry->d_inode, hidden_dentry, mode);
#else
		err = vfs_create(hidden_dir_dentry->d_inode, hidden_dentry, mode, nd);
#endif
		/* XXX this could potentially return a negative hidden_dentry! */
		if (err || !hidden_dentry->d_inode) {
			unlock_dir(hidden_dir_dentry);
			err = -EIO;
			goto end;
		}
		unlock_dir(hidden_dir_dentry);
	}

end:
	//fist_checkinode(dir, "post raif_unlink-dir");
	print_exit_status(err);
	return err;
}




STATIC int raif_unlink(inode_t *dir, dentry_t *dentry)
{
	int err = 0;

	print_entry_location();
	fist_print_dentry("IN raif_unlink: ", dentry);


	dget(dentry);

	/* delete all for large files. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		err = raif_unlink_smallfiles(dir, dentry);
#else
	err = raif_unlink_smallfiles(dir, dentry, NULL);
#endif
	

#if defined(RAIF_TAG)
	raif_rule_remove_special(dentry);
#endif

	/* call d_drop so the system "forgets" about us */
	if (!err) {
		d_drop(dentry);
	}
	

	dput(dentry);
	print_exit_status(err);
	return err;
}


STATIC int raif_rmdir_all(inode_t *dir, dentry_t *dentry) 
{
	struct dentry *hidden_dentry = NULL;
	struct dentry *hidden_dir_dentry = NULL;
	int bnum, bindex; 
	int err = 0, orig_err = 0;
	int mode = 0x666;

	print_entry_location();
	//fist_print_dentry("IN raif_rmdir_all: ", dentry);

	bnum = dbnum(dentry);
	for (bindex = 0; bindex < bnum; bindex++) {	    
		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry) {
			continue;
		}
		PASSERT(hidden_dentry);

		hidden_dir_dentry = lock_parent(hidden_dentry);

		if (IS_ERR(hidden_dir_dentry)) {
			err = PTR_ERR(hidden_dir_dentry);
			goto dec_bindex;
		}

		/* avoid destroying the hidden inode if the file is in use */
		dget(hidden_dentry);
		/* We shouldn't rmdir things in a read-only branch. */
#ifdef RAIF_TAG
//{
//{
		/* delete the .raif.rules file */
		err = raif_rule_unlink_file(hidden_dentry);
		if (err) {
			unlock_dir(hidden_dir_dentry);
			goto error;
		}
//}
//}
#endif
		mode = hidden_dentry->d_inode->i_mode;
		err = vfs_rmdir(hidden_dir_dentry->d_inode, hidden_dentry);
		fist_copy_attr_times(dir, hidden_dir_dentry->d_inode);

		/* XXX this could potentially return a negative hidden_dentry! */
		if (err || !hidden_dentry->d_inode) {
			unlock_dir(hidden_dir_dentry);
			goto error;
		}
		unlock_dir(hidden_dir_dentry);

		if (!err) {		/* since it was OK, we kill the hidden dentry. */
			dput(hidden_dentry);
			d_delete(hidden_dentry);
		}

	} // end for loop

	goto end;

dec_bindex:
	bindex--;

error:
	orig_err = err;
	// UNDO code
	// start unlinking files from bindex to 0
	for (; bindex >= 0; bindex--) {
		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry)
			continue;

		hidden_dir_dentry = lock_parent(hidden_dentry);
		err = vfs_mkdir(hidden_dir_dentry->d_inode, hidden_dentry, mode);
		unlock_dir(hidden_dir_dentry);
	}

	err = orig_err; //return original failure to userland
end:
	/* propagate number of hard-links */
	dentry->d_inode->i_nlink = get_nlinks(dentry->d_inode);

	//fist_print_dentry("OUT raif_rmdir :", dentry);
	print_exit_status(err);
	return err;
}



STATIC int raif_rmdir(inode_t *dir, dentry_t *dentry)
{
	int err = 0;

	print_entry_location();
	//fist_print_dentry("IN raif_rmdir: ", dentry);

	dget(dentry);

	/* delete all. */
	err = raif_rmdir_all(dir, dentry);
	/* call d_drop so the system "forgets" about us */
	if (!err)
		d_drop(dentry);
	dput(dentry);

	print_exit_status(err);
	return err;
}



STATIC int raif_symlink(inode_t *dir, dentry_t *dentry, const char *symname)
{
	int err = 0;
	dentry_t *hidden_dentry = NULL;
	dentry_t *hidden_dir_dentry = NULL;
	int bindex, bnum, global_bindex;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	umode_t	mode;
#endif

	print_entry_location();

	bnum = dbnum(dentry) = ibnum(dir);
	for (bindex = 0; bindex < bnum; bindex++) {	    
		hidden_dentry = dtohd_index(dentry, bindex);
		DENTRY0_TO_GLOBAL;
		if (!hidden_dentry) {
			dtohd_index(dentry, bindex) = hidden_dentry = raif_lookup_branch(dentry, bindex, NULL);
		}
		PASSERT(hidden_dentry);

		hidden_dir_dentry = lock_parent(hidden_dentry);
		if (IS_ERR(hidden_dir_dentry)) {
			unlock_dir(hidden_dir_dentry);
			err = PTR_ERR(hidden_dir_dentry);
			goto error;
		}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		err = vfs_symlink(hidden_dir_dentry->d_inode, hidden_dentry, symname);
#else
		mode = S_IALLUGO;
		err = vfs_symlink(hidden_dir_dentry->d_inode, hidden_dentry, symname, mode);
#endif
		/* XXX this could potentially return a negative hidden_dentry! */
		if (err || !hidden_dentry->d_inode) {
			unlock_dir(hidden_dir_dentry);
			goto error;
		}
		unlock_dir(hidden_dir_dentry);
	} // end for loop

	err = raif_interpose(dentry, dir->i_sb, INTERPOSE_DEFAULT);
	if (!err) {
		fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
		/* update number of links on parent directory */
		dir->i_nlink = get_nlinks(dir);
	} else {
		goto error;
	}

	goto end;

error:
	// UNDO code
	// start unlinking files from bindex -> 0
	for (; bindex >= 0; bindex--) {

		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry || !hidden_dentry->d_inode)  
			continue;

		hidden_dir_dentry = lock_parent(hidden_dentry);

		/* avoid destroying the hidden inode if the file is in use */
		dget(hidden_dentry);
		err = vfs_unlink(hidden_dir_dentry->d_inode, hidden_dentry);
		fist_copy_attr_times(dir, hidden_dir_dentry->d_inode);
		if (err) {
			unlock_dir(hidden_dir_dentry);
			goto end;
		}

		if (!err) {		/* since it was OK, we kill the hidden dentry. */
			dput(hidden_dentry);
			d_delete(hidden_dentry);
		}
		unlock_dir(hidden_dir_dentry);
	}

end:
	//fist_print_dentry("OUT raif_symlink :", dentry);
	print_exit_status(err);
	return err;
}



STATIC int raif_mkdir(inode_t *dir, dentry_t *dentry, int mode)
{
	int err = 0, orig_err = 0;
	dentry_t *hidden_dentry = NULL;
	dentry_t *hidden_dir_dentry = NULL;
	int bindex, bnum, global_bindex;

	print_entry_location();

	// Should get a negative dentry. Create on all branches.
	ASSERT(dentry->d_inode == NULL);

	bnum = dbnum(dentry) = ibnum(dir);
	for (bindex = 0; bindex < bnum; bindex++) {	    
		hidden_dentry = dtohd_index(dentry, bindex);
		DENTRY0_TO_GLOBAL;
		if (!hidden_dentry) {
			dtohd_index(dentry, bindex) = hidden_dentry = raif_lookup_branch(dentry, bindex, NULL);
		}
		PASSERT(hidden_dentry);

		hidden_dir_dentry = lock_parent(hidden_dentry);
		if (IS_ERR(hidden_dir_dentry)) {
			err = PTR_ERR(hidden_dir_dentry);
			goto dec_bindex;
		}

		err = vfs_mkdir(hidden_dir_dentry->d_inode, hidden_dentry, mode);
		/* XXX this could potentially return a negative hidden_dentry! */
		if (err || !hidden_dentry->d_inode) {
			unlock_dir(hidden_dir_dentry);
			goto error;
		}
		unlock_dir(hidden_dir_dentry);
	} // end for loop

	err = raif_interpose(dentry, dir->i_sb, INTERPOSE_DEFAULT);
	if (!err) {
		fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
		/* update number of links on parent directory */
		dir->i_nlink = get_nlinks(dir);
	} else {
		goto error;
	}

	goto end;

dec_bindex:
	bindex--;

error:
	orig_err = err;
	// UNDO code
	// start unlinking files from bindex to 0
	for (; bindex >= 0; bindex--) {

		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry || !hidden_dentry->d_inode)
			continue;

		hidden_dir_dentry = lock_parent(hidden_dentry);

		/* avoid destroying the hidden inode if the file is in use */
		dget(hidden_dentry);
		err = vfs_rmdir(hidden_dir_dentry->d_inode, hidden_dentry);
		fist_copy_attr_times(dir, hidden_dir_dentry->d_inode);
		if (err) {
			unlock_dir(hidden_dir_dentry);
			goto end;
		}

		if (!err) {		/* since it was OK, we kill the hidden dentry. */
			dput(hidden_dentry);
			d_delete(hidden_dentry);
		}
		unlock_dir(hidden_dir_dentry);
	}
	err = orig_err; //return original failure to userland
end:
	//fist_print_dentry("OUT raif_mkdir :", dentry);
	print_exit_status(err);
	return err;
}


STATIC int
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
raif_mknod(inode_t *dir, dentry_t *dentry, int mode, int dev)
#else
raif_mknod(inode_t *dir, dentry_t *dentry, int mode, dev_t dev)
#endif
{
	int err = 0;
	dentry_t *hidden_dentry = NULL;
	dentry_t *hidden_dir_dentry = NULL;
	int bnum, bindex, global_bindex;

	print_entry_location();
	//fist_print_dentry("IN raif_mknod :", dentry);

	bnum = dbnum(dentry) = ibnum(dir);
	for (bindex = 0; bindex < bnum; bindex++) {	    
		hidden_dentry = dtohd_index(dentry, bindex);
		DENTRY0_TO_GLOBAL;
		if (!hidden_dentry) {
			dtohd_index(dentry, bindex) = hidden_dentry = raif_lookup_branch(dentry, bindex, NULL);
		}
		PASSERT(hidden_dentry);

		hidden_dir_dentry = lock_parent(hidden_dentry);
		if (IS_ERR(hidden_dir_dentry)) {
			unlock_dir(hidden_dir_dentry);
			err = PTR_ERR(hidden_dir_dentry);
			goto error;
		}

		/* We shouldn't mknod things in a read-only branch. */
		err = vfs_mknod(hidden_dir_dentry->d_inode, hidden_dentry, mode, dev);
		/* XXX this could potentially return a negative hidden_dentry! */
		if (err || !hidden_dentry->d_inode) {
			unlock_dir(hidden_dir_dentry);
			goto error;
		}
		unlock_dir(hidden_dir_dentry);
	} // end for loop

	err = raif_interpose(dentry, dir->i_sb, INTERPOSE_DEFAULT);
	if (!err) {
		fist_copy_attr_timesizes(dir, hidden_dir_dentry->d_inode);
		/* update number of links on dir directory */
		dir->i_nlink = get_nlinks(dir);
	} else {
		goto error;
	}

	goto end;

error:
	//UNDO code.
	// start unlinking files from bindex -> 0
	for (; bindex >= 0; bindex--) {

		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry || !hidden_dentry->d_inode)  
			continue;

		hidden_dir_dentry = lock_parent(hidden_dentry);

		/* avoid destroying the hidden inode if the file is in use */
		dget(hidden_dentry);
		err = vfs_unlink(hidden_dir_dentry->d_inode, hidden_dentry);
		fist_copy_attr_times(dir, hidden_dir_dentry->d_inode);
		if (err) {
			unlock_dir(hidden_dir_dentry);
			goto end;
		}

		if (!err) {		/* since it was OK, we kill the hidden dentry. */
			dput(hidden_dentry);
			d_delete(hidden_dentry);
		}
		unlock_dir(hidden_dir_dentry);
	}

end:
	if (!dentry->d_inode)
		d_drop(dentry);

	//fist_print_dentry("OUT raif_mknod :", dentry);
	print_exit_status(err);
	return err;
}


STATIC int raif_rename(inode_t *old_dir, dentry_t *old_dentry,
	inode_t *new_dir, dentry_t *new_dentry)
{
	int err = 0, err1 = 0;
	dentry_t *hidden_old_dentry = NULL;
	dentry_t *hidden_new_dentry = NULL;
	dentry_t *tmp_dentry = NULL;
	dentry_t *hidden_old_dir_dentry = NULL;
	dentry_t *hidden_new_dir_dentry = NULL;
	int bnum, bindex;
	int rename_new = 0;
	int local_rule_add = 1;

	print_entry_location();


        if (!strncmp(old_dentry->d_name.name, RAIF_TMPREFIX, RAIF_TMPREFIX_LEN))
	                local_rule_add = 0;
	// XXX ARUN use old_dentry here
	// Also, if the new file exists you may need to unlink some branches
	// If it doesn't you may need to unlink one negative dentry
	bnum = dbnum(old_dentry);
	rename_new = 0;
	if (!new_dentry->d_inode) {
		//negative dentry => branch0 has the lower-level dentry
		rename_new = 1;
		tmp_dentry = dtohd_index(new_dentry, 0);
		PASSERT(tmp_dentry);
		dput(tmp_dentry);
		dtohd_index(new_dentry, 0) = NULL;
	}

	for (bindex = 0; bindex < bnum; bindex++) {	    
		//This must be changed to GLOBAL_BINDEX
		hidden_old_dentry = dtohd_index(old_dentry, bindex);
		if (rename_new) {
			dtohd_index(new_dentry, bindex) = hidden_new_dentry = raif_lookup_branch(new_dentry, GLOBAL_BINDEX(old_dentry, bindex), NULL);
		} else {
			//This must be changed to GLOBAL_BINDEX
			hidden_new_dentry = dtohd_index(new_dentry, bindex);
			if (!hidden_new_dentry) {
				dtohd_index(new_dentry, bindex) = hidden_new_dentry = raif_lookup_branch(new_dentry, GLOBAL_BINDEX(new_dentry, bindex), NULL);
			}
		}
		PASSERT(hidden_new_dentry);

		dget(hidden_old_dentry);
		dget(hidden_new_dentry);
		hidden_old_dir_dentry = get_parent(hidden_old_dentry);
		hidden_new_dir_dentry = get_parent(hidden_new_dentry);
		double_lock(hidden_old_dir_dentry, hidden_new_dir_dentry);

		err = vfs_rename(hidden_old_dir_dentry->d_inode, hidden_old_dentry,
				hidden_new_dir_dentry->d_inode, hidden_new_dentry);
		fist_dprint(5, "    raif_rename: rename branch %d status:%d\n", bindex, err);
		if (err) {
			// double_unlock will dput the new/old parent dentries whose refcnts
			// were incremented via get_parent above.
			double_unlock(hidden_old_dir_dentry, hidden_new_dir_dentry);
			dput(hidden_new_dentry);
			dput(hidden_old_dentry);
			d_drop(new_dentry);
			goto error;
		}

		// double_unlock will dput the new/old parent dentries whose refcnts
		// were incremented via get_parent above.
		double_unlock(hidden_old_dir_dentry, hidden_new_dir_dentry);
		dput(hidden_new_dentry);
		dput(hidden_old_dentry);

		fist_copy_attr_all(new_dir, hidden_new_dir_dentry->d_inode);
		if (new_dir != old_dir)
			fist_copy_attr_all(old_dir, hidden_old_dir_dentry->d_inode);
	}

	if (!err) {
		// XXX ARUN : copy attribs from src to dest
		dtopd(new_dentry)->b_num = dtopd(old_dentry)->b_num;
		dtopd(new_dentry)->mask = dtopd(old_dentry)->mask;

		if (S_ISREG(old_dentry->d_inode->i_mode)) {
#ifdef RAIF_TAG
//{
//{
			dtopd(new_dentry)->d.lookup.level = dtopd(old_dentry)->d.lookup.level;
			dtopd(new_dentry)->d.lookup.start_branch = dtopd(old_dentry)->d.lookup.start_branch;
			dtopd(new_dentry)->d.lookup.striping_unit_shift = dtopd(old_dentry)->d.lookup.striping_unit_shift;
//}
//}
#endif
		} else if (S_ISDIR(old_dentry->d_inode->i_mode)) {
#ifdef RAIF_TAG
//{
//{
			dtopd(new_dentry)->d.rules.rule = dtopd(old_dentry)->d.rules.rule;
			dtopd(new_dentry)->d.rules.local_rule = dtopd(old_dentry)->d.rules.local_rule;
			dtopd(new_dentry)->d.rules.spin = dtopd(old_dentry)->d.rules.spin;
//}
//}
#endif
		}
	}

	// XXX ARUN see if it works
	/*if (S_ISREG(old_dentry->d_inode->i_mode))*/ {
		if (new_dentry->d_inode)
			iput(new_dentry->d_inode);
		atomic_inc(&old_dentry->d_inode->i_count);
		new_dentry->d_inode = old_dentry->d_inode;
#ifdef RAIF_TAG
		if (local_rule_add) {
			err = raif_rule_rename(old_dentry, new_dentry);
			if (err)
				goto error;
		}
#else
#warning "rename operation is not implemented for no RAIF_TAG mode"
#endif
	}

	goto out;

error:
	//UNDO code
	for (bindex--; bindex >= 0; bindex--) {

		hidden_new_dentry = dtohd_index(old_dentry, bindex);

		hidden_old_dentry = lookup_one_len(old_dentry->d_name.name, hidden_new_dentry->d_parent, old_dentry->d_name.len);

		dget(hidden_new_dentry);
		hidden_old_dir_dentry = get_parent(hidden_old_dentry);
		hidden_new_dir_dentry = get_parent(hidden_new_dentry);
		double_lock(hidden_old_dir_dentry, hidden_new_dir_dentry);

		err1 = vfs_rename(hidden_old_dir_dentry->d_inode, hidden_new_dentry,
				hidden_new_dir_dentry->d_inode, hidden_old_dentry);

		if (err1) {
			// double_unlock will dput the new/old parent dentries whose refcnts
			// were incremented via get_parent above.
			double_unlock(hidden_old_dir_dentry, hidden_new_dir_dentry);
			dput(hidden_new_dentry);
			err = -EIO;
			goto out;
		}

		// double_unlock will dput the new/old parent dentries whose refcnts
		// were incremented via get_parent above.
		double_unlock(hidden_old_dir_dentry, hidden_new_dir_dentry);
		dput(hidden_new_dentry);

		fist_copy_attr_all(new_dir, hidden_new_dir_dentry->d_inode);
		if (new_dir != old_dir)
			fist_copy_attr_all(old_dir, hidden_old_dir_dentry->d_inode);
	}
out:

	fist_checkinode(new_dir, "post raif_rename-new_dir");
	print_exit_status(err);
	return err;
}

STATIC int raif_readlink(dentry_t *dentry, char *buf, int bufsiz)
{
	int err;
	dentry_t *hidden_dentry;

	print_entry_location();
	
	hidden_dentry = dtohd_index(dentry, 0);

	if (!hidden_dentry->d_inode->i_op || !hidden_dentry->d_inode->i_op->readlink) {
		err = -EINVAL;
		goto error;
	}
	err = hidden_dentry->d_inode->i_op->readlink(hidden_dentry, buf, bufsiz);
	if (err > 0)
		fist_copy_attr_atime(dentry->d_inode, hidden_dentry->d_inode);
error:
	print_exit_status(err);
	return err;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13) 
STATIC int 
#else 
STATIC void * 
#endif 
raif_follow_link(dentry_t *dentry, struct nameidata *nd)
{
	char *buf;
	int len = PAGE_SIZE, err;
	mm_segment_t old_fs;

	print_entry_location();

	buf = KMALLOC(len, GFP_KERNEL);
	if (!buf) {
		err = -ENOMEM;
		goto error;
	}

	/* read the symlink, and then we will follow it */
	old_fs = get_fs();
	set_fs(KERNEL_DS);
	err = dentry->d_inode->i_op->readlink(dentry, buf, len);
	set_fs(old_fs);
	if (err < 0)
		goto error_free;

	buf[err] = 0;	// terminate the buffer -- XXX still needed?

	/*
	 * vfs_follow_link will increase the nd's mnt refcnt
	 * we assume that some other VFS code decrements it.
	 */
	err = vfs_follow_link(nd, buf);

error_free:
	KFREE(buf);
error:
	print_exit_status(err);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13) 
	return err;
#else 
	if (err) 
		return ERR_PTR(err); 
	return NULL; 
#endif
}

/* just perform permission() on first branch */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int raif_permission(inode_t *inode, int mask, struct nameidata *nd)
#else
static int raif_permission(inode_t *inode, int mask)
#endif
{
	inode_t *hidden_inode;
	int mode;
	int err;

	print_entry_location();

	hidden_inode = itohi_index(inode, 0);
	mode = inode->i_mode;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	err = vfs_permission(inode, mask);
	if (err)
		goto error;
	err = permission(hidden_inode, mask);
#else
	err = permission(hidden_inode, mask, nd);
#endif
	print_exit_status(err);
	return err;
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
STATIC int raif_inode_revalidate(dentry_t *dentry)
{
	int err = 0;
	dentry_t *hidden_dentry;
	inode_t *hidden_inode;
	inode_t *inode;
	int bindex, bnum;

	print_entry_location();
	PASSERT(dentry);
	inode = dentry->d_inode;
	PASSERT(inode);

	bnum = dbnum(dentry);

	for (bindex = 0; bindex < bnum; bindex++) {
		hidden_dentry = dtohd_index(dentry, bindex);

		if (!hidden_dentry) {
			continue;
		}

		hidden_inode = hidden_dentry->d_inode;

		/* Now revalidate the lower level. */
		if (hidden_inode && hidden_inode->i_op && hidden_inode->i_op->revalidate) {
			err = hidden_inode->i_op->revalidate(hidden_dentry);
		}
	}

	hidden_inode = itohi_index(inode, 0);
	if (hidden_inode && !err) {
		fist_copy_attr_allbutsizes(inode, hidden_inode);
	}

	update_inode_size(inode);

	print_exit_status(err);
	return err;
}
#endif

int raif_fill_upto(file_t* hidden_file, inode_t* inodel, unsigned to_page, unsigned to);
/* TODO: If we truncate down we have to recalculate ECCs */
STATIC int raif_truncate(inode_t *inode, loff_t new_size, int bindex)
{
	int err = 0;
	inode_t* hidden_inode;
	print_entry_location();

	hidden_inode = itohi_index(inode, bindex);
	if (new_size > hidden_inode->i_size)
		fist_dprint(8, "RAIF_FILL_UPTO should be here. to page:%d, to:%d\n", INC_IF_TAG(new_size >> PAGE_CACHE_SHIFT), new_size & ~PAGE_CACHE_MASK);
	//err = raif_fill_upto(NULL, hidden_inode, INC_IF_TAG(new_size >> PAGE_CACHE_SHIFT), new_size & ~PAGE_CACHE_MASK);

	print_exit_status(err);
	return err;
}

STATIC int raif_setattr(dentry_t *dentry, struct iattr *ia)
{
	dentry_t *hidden_dentry;
	inode_t *inode = NULL;
	inode_t *hidden_inode = NULL;
	int bnum, bindex, err = 0;
	loff_t ia_size = ia->ia_size;

	print_entry_location();

	inode = dentry->d_inode;
	bnum = dbnum(dentry);
	if (ia->ia_valid & ATTR_SIZE)
		fist_dprint(8, "%s: requested size: %d\n", __FUNCTION__, ia->ia_size);

	for (bindex = 0; bindex < bnum; bindex++) {
		struct iattr temp_attr = (*ia);
		hidden_dentry = dtohd_index(dentry, bindex);
		if (!hidden_dentry) {
			continue;
		}

		if (temp_attr.ia_valid & ATTR_SIZE) {
			temp_attr.ia_size = calculate_inode_size(inode, bindex, ia_size);
			/* fill with zeros if enlarging */
			err = raif_truncate(inode, temp_attr.ia_size, bindex);
			/*if (err < 0)
			  goto out;*/
			fist_dprint(8, "%s: lower file[%d] size is: %d\n", __FUNCTION__, bindex, ia->ia_size);
		}
		err = notify_change(hidden_dentry, &temp_attr);
		if (err)
			goto out;
	}

	/* get the size from the first hidden inode */
	hidden_inode = itohi_index(inode, 0);
	fist_copy_attr_allbutsizes(inode, hidden_inode);

	/*if ((ia->ia_valid & ATTR_SIZE) && LEVEL(inode) == 0) {
	  inode->i_size = ia_size;
	//inode->i_blocks = pages;
	if ((ia_size & (PAGE_SIZE - 1)) > 0)
	pages++;
	//    inode->i_blocks++;
	// TODO: 9 should be derived from inode->i_blkbits
	//inode->i_blocks = (inode->i_blocks << (PAGE_SHIFT - 9));
	inode->i_blocks = (pages << (PAGE_SHIFT - 9));
	fist_dprint(8, "new i_size=%d, i_blocks=%d\n", inode->i_size, inode->i_blocks);
	}*/

	if (ia->ia_valid & ATTR_SIZE)
		update_inode_size(inode);
	//fist_checkinode(inode, "post raif_setattr");
out:
	print_exit_status(err);
	return err;
}

struct inode_operations raif_symlink_iops =
{
	readlink:	raif_readlink,
	follow_link:	raif_follow_link,
	permission:	raif_permission,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	revalidate:	raif_inode_revalidate,
#endif
	setattr:	raif_setattr,
};

struct inode_operations raif_dir_iops =
{
	create:		raif_create,
	lookup:		raif_lookup,
	link:		raif_link,
	unlink:		raif_unlink,
	symlink:	raif_symlink,
	mkdir:		raif_mkdir,
	rmdir:		raif_rmdir,
	mknod:		raif_mknod,
	rename:		raif_rename,
	/* no readlink/follow_link for non-symlinks */
	// off because we have setattr
	//    truncate:	raif_truncate,
	permission:	raif_permission,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	revalidate:	raif_inode_revalidate,
#endif
	setattr:	raif_setattr,
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	setxattr:	raif_setxattr,
	getxattr:	raif_getxattr,
	removexattr:	raif_removexattr,
	listxattr:	raif_listxattr,
# endif
};

struct inode_operations raif_main_iops =
{
	permission:	raif_permission,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	revalidate:	raif_inode_revalidate,
#endif
	setattr:	raif_setattr,
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	setxattr:	raif_setxattr,
	getxattr:	raif_getxattr,
	removexattr:	raif_removexattr,
	listxattr:	raif_listxattr,
# endif
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
