/*
 * Copyright (c) 2004-2006 Erez Zadok,
 * Copyright (c) 2004-2006 Nikolai Joukov,
 * Copyright (c) 2006      Chaitanya Patti,
 * Copyright (c) 2004      Abhishek Rai,
 * Copyright (c) 2004-2006 Stony Brook University
 */

#ifdef RAIF_LB

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "raif.h"
#include "loadbal.h"
#include <linux/module.h>

#include <linux/random.h>	/* for get_random_bytes */
#include <linux/sched.h>	/* for set_special_pids */

/* This definition must only appear after we include <linux/module.h> */
#ifndef MODULE_LICENSE
#define MODULE_LICENSE(foo)
#endif /* not MODULE_LICENSE */

void init_pool(struct super_block *sb) {
	stopd(sb)->poolcur = 0;
	recalc_light(sb);
}

void static inline swap(char *arr, int index1, int index2) {
	char tmp;
	if (index1 == index2)
		return;
	tmp = arr[index1];
	arr[index1] = arr[index2];
	arr[index2] = tmp;
}

void show_pool(char *pool, int end) {
	int i;
	dprintk("sbpoolend = %d\n", end); 
	dprintk("pool: ");
	for (i = 0; i <= end; i++) {
		dprintk("(%d)%d.", i, pool[i]);
	}
	dprintk("\n");
}

/* recalc_light modifies the sb structs which are also modified in 
 * branchman.c/add-branch(which uses lock_super). we can use lock_super here to increase SMP compliancy.
 * However, unless lb_fastest_branch fns also start doing that, not SMP compliant
 */
void recalc_light(struct super_block *sb) {
	int bindex;
	int min, minindex, max, maxindex;
	char *poolptr;
       	int poolend;
	char *tmppool;
	int i, j;

	/* get loads, calculate min,max */	
	minindex = maxindex = 0;
	if (minindex < 0 || maxindex < 0)
		return;
	sbload(sb, minindex) = MAX(1, lb_getload(sb, minindex));
	min = max = sbload(sb, minindex);
	for (bindex = 0 + 1; bindex < sbnum(sb); bindex++) {
		/* get load, at least 1 */
		sbload(sb, bindex) = MAX(1, lb_getload(sb, bindex));
		if (min > sbload(sb, bindex)) {
			min = sbload(sb, bindex);
			minindex = bindex;
			continue;
		}
		if (max < sbload(sb, bindex)) {
			max = sbload(sb, bindex);
			maxindex = bindex;
		}
	}

	//dprintk("Original: minindex: %d, min = %d    Maxindex: %d, Max = %d\n", minindex, min, maxindex, max);

	/* normalize */
	for (bindex = 0; bindex < sbnum(sb); bindex++) {
		sbload(sb, bindex) = MIN(MAX_RATIO, sbload(sb, bindex)/min);
	}

	/* update max, min */
	max = MIN(MAX_RATIO, max/min);	
	min = 1;

	/* Now calculate (4 * max * min / i), and create pool.
	 * max, min have the largest and smalles loads.
	 * maxindex, minindex contain corresponding indices
	 */
	poolptr = sbscratchpool(sb);
	for (bindex = 0; bindex < sbnum(sb); bindex++) {
		/* load of any branch is not more than 100 times of the min, if more, set to 100 */
		sbload(sb, bindex) = (4 * (max * min)) / sbload(sb, bindex);
		memset(poolptr, bindex, sbload(sb, bindex));
		poolptr += sbload(sb, bindex);
	}
	poolend = (poolptr - sbscratchpool(sb) -1 );

	//dprintk("<5>Changed: minindex: %d, min = %d    Maxindex: %d, Max = %d\n", maxindex, sbload(sb, maxindex), minindex, sbload(sb, minindex));
	//show_pool(sbscratchpool(sb), poolend);
	/* pool has got created in sbscratchpool(sb) and poolend
	 * Now randomize pool and swap pointers of sbpool(sb) and sbpoolend(sb)
	 */
	get_random_bytes(sbrandbuf(sb), 1 + poolend);
	for (i=0; i<poolend; i++) {
		j = sbrandbuf(sb)[i] & 0x000000ff;
		//dprintk("j = %d ", j);
		swap( sbscratchpool(sb), i, ( i + j%(poolend+1-i) ) );
	}

	/* swap the two pool pointers, they were malloc'ed equal space */
	tmppool = sbpool(sb);
	sbpool(sb) = sbscratchpool(sb);
	sbscratchpool(sb) = tmppool;

	sbpoolend(sb) = poolend;
	stopd(sb)->poolcur = 0;

	//show_pool(sbpool(sb), poolend);
}


int load_update(void *arg) {
	//struct timeval b4, af, diff;
	struct super_block *sb = (struct super_block *)arg;
	struct task_struct *tsk = current;

	/* daemonize */
	strcpy(tsk->comm, "raif_load_update");
	daemonize("raif_load_update");
	//reparent_to_init();

	while(1) {
		/* task state must be set to something b4 calling
		 * schedule_timeout (comment at schedule_...)
		 */
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(1*HZ);
		if (atomic_read(&stopd(sb)->die) == 1) {
			atomic_set(&stopd(sb)->die, 0);
			wake_up(&(stopd(sb)->clean));
			complete_and_exit(NULL, 0);
		} 
		recalc_light(sb);
	}
}



/**********************************************************
 *	Latency measuring code
 */

/* af - b4 */
static inline void tv_diff(struct timeval *af, struct timeval *b4, struct timeval *diff)
{
	diff->tv_sec = af->tv_sec - b4->tv_sec;
    
	if (af->tv_usec >= b4->tv_usec) {
		diff->tv_usec = af->tv_usec - b4->tv_usec;
	} else {
		diff->tv_sec--;
		diff->tv_usec = 1000000 + af->tv_usec - b4->tv_usec;
	}
}

struct timeval first_tv;

/* latency updating function */
void update_lat(struct timeval *last,
		oplat_t *oplat, struct timeval *b4, struct timeval *af, char *op) {
	struct timeval obs_lat_tv, current_tv;
	unsigned int obs_lat;
	static int isfirst = 1;
    
	tv_diff(af, b4, &obs_lat_tv);
	if (isfirst) {
		isfirst = 0;
		memcpy(&first_tv, af, sizeof(struct timeval));
	}
	tv_diff(af, &first_tv, &current_tv);

	obs_lat = 1000000 * (obs_lat_tv.tv_sec) + obs_lat_tv.tv_usec;
   
	memcpy(last, af, sizeof(struct timeval));

	{
		int err;
		unsigned int avg;

		avg = atomic_read(&oplat->avg);
		if (avg == 0) {
			atomic_set(&oplat->avg, obs_lat);
			return;
		}
		err = obs_lat - avg;
		if (err < (16 * oplat->dev)) {
			atomic_set(&oplat->avg, avg + (err / ALPHA));
		}
		oplat->dev = oplat->dev + ((abs(err) - oplat->dev) / BETA);
	}
}





#endif /* RAIF_LB */
